from dune.perftool.ufl.modified_terminals import Restriction
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.generation import (backend,
                                      class_member,
                                      domain,
                                      get_backend,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      include_file,
                                      kernel_cached,
                                      preamble,
                                      temporary_variable,
                                      valuearg,
                                      )
from dune.perftool.options import (get_form_option,
                                   option_switch,
                                   )
from dune.perftool.loopy.target import dtype_floatingpoint, type_floatingpoint
from dune.perftool.pdelab.quadrature import quadrature_preamble
from dune.perftool.tools import get_pymbolic_basename
from ufl.algorithms import MultiFunction
from pymbolic.primitives import Variable
from pymbolic.primitives import Expression as PymbolicExpression

from loopy.match import Writes

import numpy as np
import pymbolic.primitives as prim


@preamble
def define_reference_element(name):
    geo = name_geometry()
    include_file("dune/geometry/referenceelements.hh", filetag="operatorfile")
    return "auto {} = referenceElement({});".format(name, geo)


def name_reference_element():
    name = "refElement"
    define_reference_element(name)
    return name


@preamble
def define_localcenter(name):
    reference_element = name_reference_element()
    # Note: position(i,c) stands for the barycenter of the i-th subentity of codimension c"
    return "auto {} = {}.position(0,0);".format(name, reference_element)


def name_localcenter():
    name = "localcenter"
    define_localcenter(name)
    return name


@iname
def _component_iname(context, count):
    if context:
        context = '_' + context
    name = 'idim{}{}'.format(context, str(count))
    dim = world_dimension()
    domain(name, dim)
    return name


def component_iname(context='', count=0):
    return _component_iname(context, count)


def name_element_geometry_wrapper():
    return 'eg'


def type_element_geometry_wrapper():
    return 'EG'


def name_intersection_geometry_wrapper():
    return 'ig'


def type_intersection_geometry_wrapper():
    return 'IG'


def name_geometry_wrapper():
    """ Selects the 'native' geometry wrapper of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return name_element_geometry_wrapper()
    if it == 'exterior_facet' or it == 'interior_facet':
        return name_intersection_geometry_wrapper()
    assert False


def type_geometry_wrapper():
    """ Selects the 'native' geometry wrapper of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return type_element_geometry_wrapper()
    if it == 'exterior_facet' or it == 'interior_facet':
        return type_intersection_geometry_wrapper()
    assert False


@preamble
def define_restricted_cell(name, restriction):
    ig = name_intersection_geometry_wrapper()
    which = "inside" if restriction == Restriction.POSITIVE else "outside"
    return "const auto& {} = {}.{}();".format(name,
                                              ig,
                                              which,
                                              )


def name_cell(restriction):
    if restriction == Restriction.NONE:
        eg = name_element_geometry_wrapper()
        return "{}.entity()".format(eg)
    else:
        which = "inside" if restriction == Restriction.POSITIVE else "outside"
        name = "{}_cell".format(which)
        define_restricted_cell(name, restriction)
        return name


@preamble
def define_cell_geometry(name, restriction):
    cell = name_cell(restriction)
    return "auto {} = {}.geometry();".format(name,
                                             cell
                                             )


def name_cell_geometry(restriction):
    name = restricted_name("cell_geo", restriction)
    define_cell_geometry(name, restriction)
    return name


def type_cell_geometry(restriction):
    if restriction == Restriction.NONE:
        eg = type_element_geometry_wrapper()
        return "{}::Geometry".format(eg)
    else:
        ig = type_intersection_geometry_wrapper()
        return "{}::Entity::Geometry".format(ig)


@preamble
def define_intersection_geometry(name):
    ig = name_intersection_geometry_wrapper()
    return "auto {} = {}.geometry();".format(name,
                                             ig,
                                             )


def name_intersection_geometry():
    define_intersection_geometry("is_geo")
    return "is_geo"


def name_intersection():
    ig = name_intersection_geometry_wrapper()
    return "{}.intersection()".format(ig)


def name_geometry():
    """ Selects the 'native' geometry of the kernel """
    from dune.perftool.generation import get_global_context_value
    it = get_global_context_value("integral_type")

    if it == 'cell':
        return name_cell_geometry(Restriction.NONE)
    if it == 'exterior_facet' or it == 'interior_facet':
        return name_intersection_geometry()
    assert False


@preamble
def define_in_cell_geometry(restriction, name):
    ig = name_intersection_geometry_wrapper()
    which = "In" if restriction == Restriction.POSITIVE else "Out"
    return "auto {} = {}.geometryIn{}side();".format(name,
                                                     ig,
                                                     which
                                                     )


def name_in_cell_geometry(restriction):
    assert restriction is not Restriction.NONE

    name = "geo_in_{}side".format("in" if restriction is Restriction.POSITIVE else "out")
    define_in_cell_geometry(restriction, name)
    return name


# TODO is it always necessary to add quadrature inames?
def apply_in_cell_transformation(name, local, restriction):
    geo = name_in_cell_geometry(restriction)
    return quadrature_preamble("{} = {}.global({});".format(name,
                                                            geo,
                                                            str(local),
                                                            ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(local)}),
                               depends_on=frozenset({Writes(get_pymbolic_basename(local))}),
                               )


def pymbolic_in_cell_coordinates(local, restriction):
    basename = get_pymbolic_basename(local)
    name = "{}_in_{}side".format(basename, "in" if restriction is Restriction.POSITIVE else "out")
    temporary_variable(name, shape=(world_dimension(),), shape_impl=("fv",))
    apply_in_cell_transformation(name, local, restriction)
    return Variable(name)


def to_cell_coordinates(local, restriction):
    assert isinstance(local, PymbolicExpression)
    if restriction == Restriction.NONE:
        return local
    else:
        return pymbolic_in_cell_coordinates(local, restriction)


def world_dimension():
    data = get_global_context_value("data")
    form = data.object_by_name[get_form_option("form")]
    from dune.perftool.ufl.preprocess import preprocess_form
    form = preprocess_form(form).preprocessed_form
    return form.ufl_cell().geometric_dimension()


def intersection_dimension():
    return world_dimension() - 1


def local_dimension():
    it = get_global_context_value('integral_type')
    if it == "cell":
        return world_dimension()
    else:
        return intersection_dimension()


def evaluate_unit_outer_normal(name):
    ig = name_intersection_geometry_wrapper()
    qp = get_backend("quad_pos")()
    return quadrature_preamble("{} = {}.unitOuterNormal({});".format(name, ig, qp),
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(qp)}),
                               depends_on=frozenset({Writes(get_pymbolic_basename(qp))}),
                               )


@preamble
def declare_normal(name, kernel, decl_info):
    ig = name_intersection_geometry_wrapper()
    return "auto {} = {}.centerUnitOuterNormal();".format(name, ig)


def pymbolic_unit_outer_normal():
    name = "unit_outer_normal"
    if not get_form_option("diagonal_transformation_matrix"):
        temporary_variable(name, shape=(world_dimension(),), decl_method=declare_normal)
        evaluate_unit_outer_normal(name)
    else:
        declare_normal(name, None, None)
        globalarg(name, shape=(world_dimension(),))
    return prim.Variable(name)


def evaluate_unit_inner_normal(name):
    outer = name_unit_outer_normal()
    return quadrature_preamble("auto {} = -1. * {};".format(name, outer),
                               assignees=frozenset({name}),
                               read_variables=frozenset({outer}),
                               )


def pymbolic_unit_inner_normal():
    name = "inner_normal"
    temporary_variable(name, shape=(world_dimension(),), decl_method=declare_normal)
    evaluate_unit_inner_normal(name)
    return prim.Variable(name)


def type_jacobian_inverse_transposed(restriction):
    geo = type_cell_geometry(restriction)
    return "typename {}::JacobianInverseTransposed".format(geo)


@kernel_cached
def define_jacobian_inverse_transposed_temporary(restriction):
    @preamble
    def _define_jacobian_inverse_transposed_temporary(name, kernel, decl_info):
        t = type_jacobian_inverse_transposed(restriction)
        return "{} {};".format(t,
                               name,
                               )
    return _define_jacobian_inverse_transposed_temporary


@preamble(kernel="operator")
def define_constant_jacobian_inverse_transposed_eval(name):
    from dune.perftool.pdelab.localoperator import name_ansatz_gfs_constructor_param, lop_template_range_field
    gfs = name_ansatz_gfs_constructor_param()
    rft = lop_template_range_field()

    return "{} = {}.gridView().template begin<0>()->geometry().jacobianInverseTransposed(Dune::FieldVector<{}, {}>());".format(name, gfs, rft, world_dimension())


@class_member(classtag="operator")
def define_constant_jacobian_inverse_transposed(name):
    define_constant_jacobian_inverse_transposed_eval(name)
    from dune.perftool.pdelab.localoperator import lop_template_ansatz_gfs
    gfst = lop_template_ansatz_gfs()
    return "typename {}::Traits::GridView::template Codim<0>::Geometry::JacobianInverseTransposed {};".format(gfst, name)


@backend(interface="name_jit", name="constant_transformation_matrix")
def name_constant_jacobian_inverse_transposed(restriction):
    name = "jit"
    dim = world_dimension()
    globalarg(name, shape=(dim, dim), managed=False)
    define_constant_jacobian_inverse_transposed(name)
    return name


def define_jacobian_inverse_transposed(name, restriction):
    dim = world_dimension()
    temporary_variable(name, decl_method=define_jacobian_inverse_transposed_temporary(restriction), shape=(dim, dim))
    geo = name_cell_geometry(restriction)
    pos = get_backend("qp_in_cell", selector=option_switch("blockstructured"))(restriction)

    return quadrature_preamble("{} = {}.jacobianInverseTransposed({});".format(name,
                                                                               geo,
                                                                               str(pos),
                                                                               ),
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(pos)}),
                               depends_on=frozenset({Writes(get_pymbolic_basename(pos))}),
                               )


@backend(interface="name_jit", name="default")
def name_jacobian_inverse_transposed(restriction):
    name = restricted_name("jit", restriction)
    define_jacobian_inverse_transposed(name, restriction)
    return name


def pymbolic_jacobian_inverse(i, j, restriction):
    # Calculate jacobian_inverse_transposed
    name = get_backend(interface="name_jit", selector=option_switch("constant_transformation_matrix"))(restriction)

    # Return jacobian inverse -> (j, i) instead of (i, j)
    return prim.Subscript(prim.Variable(name), (j, i))


@preamble(kernel="operator")
def define_constant_jacobian_determinant_eval(name):
    from dune.perftool.pdelab.localoperator import name_ansatz_gfs_constructor_param, lop_template_range_field
    gfs = name_ansatz_gfs_constructor_param()
    rft = lop_template_range_field()

    return "{} = {}.gridView().template begin<0>()->geometry().integrationElement(Dune::FieldVector<{}, {}>());".format(name, gfs, rft, world_dimension())


@class_member(classtag="operator")
def _define_constant_jacobian_determinant(name):
    define_constant_jacobian_determinant_eval(name)
    from dune.perftool.pdelab.localoperator import lop_template_ansatz_gfs
    gfst = lop_template_ansatz_gfs()
    return "typename {}::Traits::GridView::template Codim<0>::Geometry::ctype {};".format(gfst, name)


@backend(interface="detjac", name="constant_transformation_matrix")
def define_constant_jacobian_determinant(name):
    valuearg(name)
    _define_constant_jacobian_determinant(name)


@backend(interface="detjac", name="default")
def define_jacobian_determinant(name):
    temporary_variable(name, shape=())
    geo = name_geometry()
    pos = get_backend("quad_pos", selector=option_switch("blockstructured"))()
    code = "{} = {}.integrationElement({});".format(name,
                                                    geo,
                                                    str(pos),
                                                    )

    return quadrature_preamble(code,
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(pos)}),
                               depends_on=frozenset({Writes(get_pymbolic_basename(pos))}),
                               )


@backend(interface="fdetjac", name="constant_transformation_matrix")
@preamble
def define_facet_jacobian_determinant(name):
    # NB: This might be optimized to store *d* values on the operator level
    #     We don't do that right now for laziness.
    geo = name_geometry()
    pos = name_localcenter()

    valuearg(name)

    return "auto {} = {}.integrationElement({});".format(name,
                                                         geo,
                                                         pos,
                                                         )


@backend(interface="fdetjac", name="default")
def define_facet_jacobian_determinant(name):
    return define_jacobian_determinant(name)


def name_jacobian_determinant():
    name = 'detjac'
    get_backend(interface="detjac", selector=option_switch("constant_transformation_matrix"))(name)
    return name


def pymbolic_jacobian_determinant():
    return prim.Variable(name_jacobian_determinant())


def name_facet_jacobian_determinant():
    name = 'fdetjac'
    get_backend(interface="fdetjac", selector=option_switch("constant_transformation_matrix"))(name)
    return name


def pymbolic_facet_jacobian_determinant():
    return prim.Variable(name_facet_jacobian_determinant())


def apply_to_global_transformation(name, local):
    temporary_variable(name, shape=(world_dimension(),), shape_impl=("fv",))
    geo = name_geometry()
    code = "{} = {}.global({});".format(name,
                                        geo,
                                        local,
                                        )
    return quadrature_preamble(code,
                               assignees=frozenset({name}),
                               read_variables=frozenset({get_pymbolic_basename(local)}),
                               depends_on=frozenset({Writes(get_pymbolic_basename(local))})
                               )


def to_global(local):
    assert isinstance(local, prim.Expression)
    name = get_pymbolic_basename(local) + "_global"
    apply_to_global_transformation(name, local)
    return prim.Variable(name)


@preamble
def define_cell_volume(name, restriction):
    geo = name_cell_geometry(restriction)
    valuearg(name)
    return "auto {} = {}.volume();".format(name, geo)


def pymbolic_cell_volume(restriction):
    if get_form_option("constant_transformation_matrix"):
        return pymbolic_jacobian_determinant()
    else:
        name = restricted_name("volume", restriction)
        define_cell_volume(name, restriction)
        return prim.Variable(name)


@preamble
def define_facet_area(name):
    geo = name_intersection_geometry()
    valuearg(name)
    return "auto {} = {}.volume();".format(name, geo)


def pymbolic_facet_area():
    if get_form_option("constant_transformation_matrix"):
        return pymbolic_facet_jacobian_determinant()
    else:
        name = "area"
        define_facet_area(name)
        return prim.Variable(name)


def define_element_corners(name):
    from dune.perftool.pdelab.driver import get_form
    n_corners = get_form().ufl_cell().num_vertices()
    temporary_variable(name, shape_impl = ('fv', 'fv'), shape = (n_corners, local_dimension()))



def name_element_corners():
    name = "corners"
    define_element_corners(name)
    return name